/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2002 Kamil Skalski.
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 703 $ $LastChangedDate: 2002-08-30 01:19:46 +0200 (Fri, 30 Aug 2002) $
 *
 * Patrange - check for pattern exhaustiveness and unused cases.
 *
 * The algorithm used here was developed by Kamil Skalski. I just recoded
 * it from OCaml to Gont. However the generate example code in old compiler
 * was wrong, for example for:
 *
 *	switch (x, y) {
 *	case (true, false):
 *	case (false, true):
 *	case (true, true):
 *	}
 * 
 * it generates: (_, false). However as I was unable to figure out how
 * to do it The Right Way, I use randomized pseudo-algorithm of my own,
 * to generate example.
 */

// TODO: fix generate example not to be random ;)

// someday-TODO: what to do:
// switch ex {
// case Foo: 
// 	// ...
// case Baar: 
//	// ...
// }
// where Baar is spelling error? The problem isn't as big in switch, but
// it is in with { ... } matching. -Wcapitalization has been now implemented,
// that warns about capitalized pattern variables. We might want 
// -Wstrict-capitalization, that warns about non-capitalize constructors etc.
// -Wcapitalization should only warn about potentially dangerous things.

// someday-TODO: unused clause checks for with-matching

open Ast;
open Util;
open Tyutil;

exception void No_example;
exception void Break_loop;

union range {
	void Omega;	// match-all, `_' or `x'
	bool BoolRange;
	<<range>list>list TupleRange;
	*(var_symbol, <range>list) UnionRange;
	expr ExprRange;
}

int expr_id;


string string_of_range(range r)
{
	string xalternatives(<range>list l)
	{
		let sl = List::qsort(String::cmp, 
				     List::map(string_of_range, l)) in
		return String::concat("|", sl);
	}
	
	switch r {
	case Omega:		return "_";
	case BoolRange(true):	return "T";
	case BoolRange(false):	return "F";
	case ExprRange(_):	return "X<" + itoa(expr_id++) + ">";
		
	case TupleRange(lst):
		return "(" + 
			String::concat(", ", List::map(xalternatives, lst)) + 
		       ")";
		       
	case UnionRange(s, lst):
		return symbol_name(s) + "(" + xalternatives(lst) + ")";
	}
}

string alternatives(<range>list l)
{
	let sl = List::qsort(String::cmp, 
			     List::map(string_of_range, l)) in
	return String::concat(" | ", sl);
}
	

string string_of_example(range r)
{
	string just_one(<range>list lst)
	{
		switch lst {
		case Cons(x, Nil):
			return string_of_example(x);
		case _:
			return ice("Patrange::string_of_example(): non-1");
		}
	}
	
	switch r {
	case Omega:		return "_";
	case BoolRange(true):	return "true";
	case BoolRange(false):	return "false";
	case ExprRange(e):	return "some " + string_of_type(type_of(e));
		
	case TupleRange(lst):
		return "(" + 
			String::concat(", ", List::map(just_one, lst)) + 
		       ")";
		       
	case UnionRange(s, lst):
		return symbol_name(s) + "(" + just_one(lst) + ")";
	}
}

// check whatever big >= small
bool is_subrange_of(range big, range small)
{
	switch (big, small) {
	case (Omega, _):
		return true;
		
	case (BoolRange(x), BoolRange(y)):
		return x == y;
		
	case (TupleRange(l1), TupleRange(l2)):
		void cross_check(<range>list b, <range>list s)
		{
			bool check_one(range ss)
			{
				let f = fun bool (range bb) 
						(is_subrange_of(bb, ss)) in
				return !List::exists(f, b);
			}

			if (List::exists(check_one, s))
				raise Break_loop;
		}

		try {
			List::iter2(cross_check, l1, l2);
			return true;
		} with {
		case Break_loop:
			return false;
		}
		
	case (UnionRange(s1, rr1), UnionRange(s2, r2)):
		<range>list r1 = rr1;

		if (!same_symbols(s1, s2))
			return false;

		bool check_one(range s)
		{
			return !List::exists(fun bool (range b) 
							(is_subrange_of(b, s)),
					     r1);
		}

		return !List::exists(check_one, r2);
	
	case (_, _):
		return false;
	}
}

// speed-TODO: rewrite
bool is_same_range(range r1, range r2)
{
	return string_of_range(r1) == string_of_range(r2);
}

// likewise
bool same_alternatives(<range>list l1, <range>list l2)
{
	if (List::length(l1) != List::length(l2))
		return false;
	
	let s1 = List::qsort(String::cmp, List::map(string_of_range, l1)) in
	let s2 = List::qsort(String::cmp, List::map(string_of_range, l2)) in
	return String::concat("@", s1) == String::concat("@", s2);
}

range xfold(range r)
{
	switch r {
	case TupleRange(lst):
		bool isnt_ok(<range>list l)
		{
			switch l {
			case Cons(x, Nil):
				switch xfold(x) {
				case Omega:
					return false;
				case _:
					return true;
				}
			case _:
				return true;
			}
		}

		if (List::exists(isnt_ok, lst))
			return r;
		else
			return Omega;

	case UnionRange(s, Cons(x, Nil)):
		return UnionRange(s, List::cons1(xfold(x)));

	case _:
		return r;
	}
}

range fold(range r)
{
//dbg("(fold: " + string_of_range(r) + " -> ");
	r = xfold(r);
//dbg(string_of_range(r) + ")\n");
	return r;
}

range range_of_pattern(pattern p)
{
	range r;

	switch p.a {
	case VoidPattern:
		return Omega;
		
	case VarPattern(v):
		return Omega;
	
	case ExprPattern(e):
		switch e.a.raw {
		case BoolConst(x):
			return BoolRange(x);
		case _:
			return ExprRange(e);
		}

		
	case TuplePattern(lst):
		<range>list f(pattern pp)
		{
			return List::cons1(range_of_pattern(pp));
		}
		
		r = TupleRange(List::map(f, lst));
	
	case UnionPattern(np):
		let x = range_of_pattern(np.np_pattern) in
		r = UnionRange(get_binding(np.np_symbol), List::cons1(x));
	}

	return fold(r);
}


<var_symbol>list union_fields(var_symbol one_field)
{
	var_symbol get_fld_name(<var_symbol>field f)
	{
		return get_binding(f.fld_symbol);
	}
	
	switch (union_of(one_field)).stype {
	case Union(x):
		return List::map(get_fld_name, x);
	case _:
		return ice("Patrange::union_fields(): " 
			   + "bad parent union");
	}
}


<range>list union_check(<range>list lst)
{
	<var_symbol>list fields;
	
//dbg("(union_check: " + alternatives(lst) + "\n");

	switch lst {
	case Cons(UnionRange(s, _), _):
		fields = union_fields(s);
		
	case _:
		return lst;
	}

	if (List::length(lst) != List::length(fields)) {
//dbg("union_check: bad ret)\n");
		return lst;
	}

	bool isnt_in_lst(var_symbol s)
	{
		bool is_ok(range r)
		{
			switch r {
			case UnionRange(ss, Cons(Omega, Nil)):
				return same_symbols(s, ss);
				
			case UnionRange(_, _):
				return false;
			
			case _:
				ice("Patrange::union_check(): " + 
					   "non-UnionRange");
				return false;
			}
		}

		return !List::exists(is_ok, lst);
	}

	if (!List::exists(isnt_in_lst, fields))
		lst = List::cons1(Omega);
//dbg("union_check_ret: " + alternatives(lst) + ")\n");
	return lst;
}

<range>list merge_in(<range>list lst, range new)
{
	new = fold(new);

	if (List::exists(fun bool (range x) (is_subrange_of(x, new)), lst)) {
//dbg("warn: " + string_of_range(new) + " to " + alternatives(lst) + "\n");
		warn("unused pattern");
		return lst;
	}
	
//dbg("(merge_in: " + string_of_range(new) + " to " + alternatives(lst) + "\n");
	<range>option find_place_in(range old)
	{
		switch (old, new) {
		case (_, Omega):
			return Some(Omega);
		
		case (UnionRange(s1, l1), UnionRange(s2, l2)):
			if (!same_symbols(s1, s2))
				return None;
			
			let n = List::fold_left(merge_in, l1, l2) in
			return Some(UnionRange(s1, n));
	
		case (BoolRange(true), BoolRange(false)):
			return Some(Omega);
			
		case (BoolRange(false), BoolRange(true)):
			return Some(Omega);
		
		case (TupleRange(l1), TupleRange(l2)):
			<<range>list>list ret = Nil;
			int pos = 0, diff_at = -1;

			void chk(<range>list x, <range>list y)
			{
				if (!same_alternatives(x, y)) {
					if (diff_at != -1)
						raise Break_loop;
					else
						diff_at = pos;
				}

				pos++;
			}

			void merge_em(<range>list x, <range>list y)
			{
				if (pos++ == diff_at)
					x = List::fold_left(merge_in, x, y);
				
				ret = Cons(x, ret);
			}


			try {
				List::iter2(chk, l1, l2);
				if (diff_at == -1)
					ice("Patrange::merge_in(): the same");
				pos = 0;
				List::iter2(merge_em, l1, l2);
				return Some(fold(TupleRange(List::rev(ret))));
			} with {
			case Break_loop:
				return None;
			}
			
		case (_, _):
			return None;
		}
	}

	<range>option merged = None;
	
	<range>list iter(<range>list a)
	{
		switch a {
		case Nil:
			return List::cons1(new);
			
		case Cons(x, xs):
			switch find_place_in(x) {
			case Some(y):
				merged = Some(y);
				return xs;
			case None:
				if (is_subrange_of(new, x))
					return iter(xs);
				else
					return Cons(x, iter(xs));
			}
		}
	}

	lst = iter(lst);

	switch merged {
	case Some(x):
		new = x;
		bool aint_subrange(range r)
		{
			return !is_subrange_of(new, r);
		}
	
		lst = merge_in(List::filter(aint_subrange, lst), new);
		
	case None:
		skip;
	}
	
	lst = union_check(lst);
//dbg("merge_in_ret: " + alternatives(lst) + ")\n");
	return lst;
}

void check_exhaustiveness(<*(pattern, bool)>list pats)
{
	<range>list ctx = Nil;

	if (!flags.wpattern)
		return;
	
	void check_one(*(pattern, bool) pp)
	{
		let (pat, can_fail) = pp in {
			push_location(pat.loc);
			let r = range_of_pattern(pat) in {
				if (can_fail)
					merge_in(ctx, r);
				else
					ctx = merge_in(ctx, r);
			}
			pop_location();
		}
	}

	List::iter(check_one, pats);

	switch ctx {
	case Cons(Omega, Nil):
		return;		// ok
	case _:
		warn("this pattern matching is not exhaustive");

		try {
			string ex = string_of_example(generate_example(ctx));
			warn("example of value that isn't matched: " + ex);
		} with {
		case No_example:
			warn("cannot generate example in resoanable time, " +
			     "compiler's author is lame");
		}
	}
}

// check whether pat could match example
bool could_match_1(range pat, range example)
{
	switch (pat, example) {
	case (Omega, _):
		return true;
		
	case (_, Omega):
		return true;
		
	case (BoolRange(x), BoolRange(y)):
		return x == y;
		
	case (TupleRange(l1), TupleRange(l2)):
		void cross_check(<range>list pat_l, <range>list example_l)
		{
			range examp;
			
			switch example_l {
			case Cons(x, Nil):
				examp = x;
			case _:
				ice("Patrange::could_match_1(): non-1");
				return;
			}
			
			if (!List::exists(fun bool (range pa) 
						(could_match_1(pa, examp)), 
					  pat_l))
				raise Break_loop;
		}

		try {
			List::iter2(cross_check, l1, l2);
			return true;
		} with {
		case Break_loop:
			return false;
		}
		
	case (UnionRange(s1, pat_l), UnionRange(s2, ex_l)):
		if (!same_symbols(s1, s2))
			return false;

		switch ex_l {
		case Cons(exx, Nil):
			range ex = exx;
			return List::exists(fun bool (range p) 
							(could_match_1(p, ex)),
					    pat_l);

		case _:
			ice("Patrange::could_match_1(): non-1?");
			return false;
		}

	case (_, _):
		return false;
	}
}

bool could_match(<range>list pat_l, range ex)
{
	return List::exists(fun bool (range p) (could_match_1(p, ex)), pat_l);
}

int rand_seed;

int rand()
{
	// but these are taken from Bruce Schneier: Applied Cryptography
	rand_seed = (rand_seed * 4096 + 150889) % 714024;

	return rand_seed;
}

'a pick_random_from_list(<'a>list lst)
{
	int n = List::length(lst);

	if (n == 0)
		ice("Patrange::pick_random_from_list(): n = 0");

	n = rand() % n;

	while (n-- > 0)
		switch lst {
		case Cons(_, tl):
			lst = tl;
		case Nil:
			ice("Patrange::pick_random_from_list(): impossible");
		}
		
	switch lst {
	case Cons(hd, _):
		return hd;
	case Nil:
		return ice("Patrange::pick_random_from_list(): impossible 2");
	}
}

range union_example(var_symbol us, <range>list ranges)
{
	<var_symbol, <range>list>Ptrhash::t matched = 
					Ptrhash::create(13);
	
	void add(range r)
	{
		switch r {
		case UnionRange(s, l):
			Ptrhash::add(matched, s, l);
		case _:
			ice("Patrange::union_example(): non-union");
		}
	}

	List::iter(add, ranges);

	<range>list ret = Nil;

	void check(var_symbol s)
	{
		if (Ptrhash::mem(matched, s))
			let rs = Ptrhash::find(matched, s) in
			try {
				let ex = List::cons1(generate_example(rs)) in
				ret = Cons(UnionRange(s, ex), ret);
			} with {
			case No_example:
				skip;
			}
		else
			let p = UnionRange(s, List::cons1(Omega)) in
			ret = Cons(p, ret);
	}

	List::iter(check, union_fields(us));

	return pick_random_from_list(ret);
}

range tuple_example(<<range>list>list lst, <range>list rest)
{
	<range>list examples = Nil;

	<<range>list>list h1(<<range>list>list so_far,
				   <<range>list>list out_of)
	{
		switch out_of {
		case Cons(hd, tl):
			let rnd = pick_random_from_list(hd) in
			let rec = h1(Cons(hd, so_far), tl) in
			let ret = Cons(List::cons1(rnd), rec) in
			try {
				let ex = List::cons1(generate_example(hd)) in
				let exa = TupleRange(
					List::concat(List::rev(so_far),
					Cons(ex, rec))) in
				examples = Cons(exa, examples);
				return ret;
			} with {
			case No_example:
				return ret;
			}

		case Nil:
			return Nil;
		}
	}

	bool try_it_out(range ex)
	{
//	dbg("ex: " + string_of_range(ex) + "\n");
		return !could_match(rest, ex);
	}
	
	// try few times
	for (int i = 0; i < 32; i++) {
//dbg("i="+itoa(i)+"\n");
		h1(Nil, lst);
		try {
			return List::find(try_it_out, examples);
		} with {
		case Not_found:
			skip;
		}
		examples = Nil;
	}

	raise No_example;
}

// Generate random example. We will stick to it until we'll get something
// more reasonable.
range generate_example(<range>list ranges)
{
	void check_nil(<range>list nil)
	{
		switch nil {
		case Nil:
			skip;
		case _:
			ice("Patrange::generate_example(): should be 1-long: " +
				alternatives(ranges));
		}
	}
	
	switch ranges {
	case Cons(BoolRange(x), nil):
		// there cannot be BoolRange(!x) in list
		check_nil(nil);
		return BoolRange(!x);
		
	case Cons(Omega, nil):
		check_nil(nil);
		raise No_example;
	
	case Cons(ExprRange(e), lst):
		bool aint_expr(range r)
		{
			switch r {
			case ExprRange(_):	return false;
			case _:			return true;
			}
		}
		
		check_nil(List::filter(aint_expr, lst));

		return ExprRange(e);

	case Cons(UnionRange(us, _), _):
		return union_example(us, ranges);
	
	case Cons(TupleRange(lst), rest):
		return tuple_example(lst, rest);
	
	case Nil:
		return ice("Patrange::generate_example(): nil");
	}
}

