/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 693 $ $LastChangedDate: 2002-08-29 23:01:04 +0200 (Thu, 29 Aug 2002) $
 *
 * Type - Type checker.
 */

open Ast;
open Util;
open Tyutil;

/* Raised by routines below in case of problem. Should not get out this 
   module.  */
exception string Type_error;

struct ty_ctx_desc {
    ty_ctx tctx;
}

struct ty_ctx {
    <xtyvar, ty_ctx_desc>Unionfind::t var_uses;
    ty fnc_ret_type;
    bool seen_return_with_expression;
    int level;
    <ty_ctx>option parent;
}

ty_ctx dummy_ty_ctx()
{
    return {var_uses = Unionfind::create(),
	    fnc_ret_type = AbstractType,
	    parent = None,
	    level = 0,
	    seen_return_with_expression = false};
}

/* Same as type_instanace() except that it is safe to call on types with
   weak variables.  */
ty xt_instance(ty t)
{
    t = normalize_xt(t);
    
    switch t {
    case XTyVar(_):
	skip;
    case _:
	t = type_instance(t);
    }

    return t;
}

void save_fv(ty_ctx c, xtyvar x)
{
    let desc = {tctx = c} in {
        Unionfind::make_set(c.var_uses, desc);
	Unionfind::insert(c.var_uses, x, desc);
    }
}

void union_vars(ty_ctx c, xtyvar v1, xtyvar v2)
{
    let d1 = Unionfind::find(c.var_uses, v1) in
    let d2 = Unionfind::find(c.var_uses, v2) in {
	if (d1.tctx.level == d2.tctx.level &&
	    !Hacks::same_ptrs(d1.tctx, d2.tctx))
	    ice("Type::union_vars(): different, same level contexts");
	    
	if (d1.tctx.level < d2.tctx.level)
	    Unionfind::union_sets(c.var_uses, d1, d2);
	else
	    Unionfind::union_sets(c.var_uses, d2, d1);
    }
}

ty fv(ty_ctx c, ty t, <id>list vars)
{
    let (lst, tt) = subst_xtyvars(t, vars) in {
	List::iter(partial(save_fv, c), lst);
	return tt;
    }
}

void push_fv(ty_ctx c, ty t)
{
    switch xt_instance(t) {
    case XTyVar(x): save_fv(c, x);
    case _:	    skip;
    }
}

/* Walk down .t chain until we'll get to real type or None.  */
xtyvar walk_down(xtyvar v)
{
    switch v.t {
    case Some(XTyVar(x)):
	return walk_down(x);
    case Some(_):
    case None:
	return v;
    }
}

/* v <- t1 */
void subst(ty_ctx c, xtyvar v, ty t1)
{
    void f(ty t)
    {
	switch t {
	case XTyVar(x):
	    switch x.t {
	    case None:
		union_vars(c, v, x);
	    case Some(_):
		skip;
	    }
	case _:
	    skip;
	}
    }
    
//dbgl("Subst: " + v.name + " <- " + string_of_type(t1));
    type_iter(f, t1);
    
    switch v.t {
    case None:
	v.t = Some(t1);
    case Some(_):
	ice("Type::subst(): already done");
    }
}

void unify(ty_ctx c, ty t1, ty t2)
{
    void bail_out(string m)
    {
	raise Type_error("cannot unify " + string_of_type(t1) + " with " + 
			 string_of_type(t2) + m);
    }

    void self(_ x, _ y)
    {
	unify(c, x, y);
    }

//dbgl("UN: " + string_of_type(t1) + " <-> " + string_of_type(t2));
    
    switch (t1, t2) {
    case (Int, Int):
    case (Void, Void):
    case (Bool, Bool):
    case (Float, Float):   
    case (String, String):
    case (Exn, Exn):
	skip;

    case (Tuple(l1), Tuple(l2)):
	if (List::length(l1) != List::length(l2))
	    bail_out(" because they have different number of arguments");
	List::iter2(self, l1, l2);
    
    case (Func(ft1), Func(ft2)):
	switch (fv(c, t1, ft1.tyvars), fv(c, t2, ft2.tyvars)) {
	case (Func(f1), Func(f2)):
	    if (List::length(f1.parms) != List::length(f2.parms))
		bail_out(" because they have different number of arguments");
	    self(f1.ret_type, f2.ret_type);
	    // someday-TODO: include name of parameter in error message
	    List::iter2(self, List::map(fst, f1.parms), 
			List::map(fst, f2.parms));
	case _:
	    ice("Type::unify(): func then non func");
	}
	
    case (TyVar(v1), TyVar(v2)):
	if (v1 != v2)
	    raise Type_error("cannot unify type variables '" + v1 + 
			     " and '" + v2);
    
    case (XTyVar(xx1), XTyVar(xx2)):
	_ x1 = walk_down(xx1);
	_ x2 = walk_down(xx2);


	// if they are the same, then there is nothing left to do
	if (!Hacks::same_ptrs(x1, x2)) {
	    // otherwise the funny part starts ;)
	    
	    switch (x1.t, x2.t, x1, x2) {
	    case (Some(tt1), Some(tt2), _, _):
		self(tt1, tt2);
		
	    case (Some(t), None, _, x):
	    case (None, Some(t), x, _):
		subst(c, x, t);
		
	    case (None, None, _, _):
		subst(c, x1, XTyVar(x2));
	    }

	}
	    
    case (t, XTyVar(xx)):
    case (XTyVar(xx), t):
	_ x = walk_down(xx);

	switch x.t {
	case Some(s):
	    self(s, t);
	    
	case None:
	    subst(c, x, t);
	}

    // below this point neither [t1] nor [t2] is [XTyVar(...)] so we might
    // safely use is_type_instance(), is_base_type() etc.
	
    // opt_struct {} is type of null
    case (OptStruct(Nil), OptStruct(_)): 
    case (OptStruct(_), OptStruct(Nil)): 
	skip;
    
    case (OptStruct(Nil), _):
	if (is_type_instance(t2))
	    bail_out("");
	self(t1, type_instance(t2));
	
    case (_, OptStruct(Nil)):
	if (is_type_instance(t1))
	    bail_out("");
	self(type_instance(t1), t2);

    // named-type/base-type stuff
    case (NamedType(nt1), NamedType(nt2)):
	bool ss = same_symbols(get_binding(nt1.tydef), get_binding(nt2.tydef));

	if (ss) {
	    List::iter2(self, nt1.parms, nt2.parms);
	} else {
	    if (!is_base_type(t1) || !is_base_type(t2)) {
		self(base_type(t1), base_type(t2));
	    } else {
		bail_out("");
	    }
	}
    
    case (NamedType(_), _):
	if (!is_base_type(t1))
	    self(base_type(t1), t2);
	else
	    bail_out("");

    case (_, NamedType(_)):
	if (!is_base_type(t2))
	    self(t1, base_type(t2));
	else
	    bail_out("");
    
    case (AbstractType, AbstractType):
	// slightly better error message then below
	raise Type_error("cannot unify different abstract types");
	
    case (_, _):
	bail_out("");
    }
}

void tycheck_lambda(<ty_ctx>option c, lambda_record ll)
{
    _ uses, lev;
    
    switch c {
    case None:
	uses = Unionfind::create();
	lev = 1;
	
    case Some(p):
	uses = p.var_uses;
	lev = p.level + 1;
    }
    
    let ft = get_ft_record(ll.lambda_type) in
    let cc = {fnc_ret_type = ft.ret_type, 
	      var_uses = uses,
	      level = lev,
	      parent = c,
	      seen_return_with_expression = false} in 
    {
	int tyvar_cnt = 0;
	<id>list free_vars = Nil;

	void subst_fv(ty t)
	{
//dbgl("sv: " + string_of_type(t));
	    switch t {
	    case XTyVar(x):
		switch x.t {
		case None:
//dbgl("save " + x.name);
		    let d = Unionfind::find(cc.var_uses, x) in
		    if (d.tctx.level == cc.level) {
			if (!Hacks::same_ptrs(d.tctx, cc))
			    ice("Type::tycheck_lambda(): same level, "
				"different contexts");
			let n = "v" + itoa(tyvar_cnt++) in {
			    free_vars = Cons(n, free_vars);
			    x.t = Some(TyVar(n));
//dbgl("as " + n);
			}
		    } else {
//dbgl("lev = " + itoa(d.tctx.level) + " our = " + itoa(cc.level));
		    }

		case Some(_):
		    skip;
		}

	    case _:
		skip;
	    }
	}
 
	List::iter(fun (a) (push_fv(cc, fst(a))), ft.parms);
	push_fv(cc, ft.ret_type);

	tycheck_stmt(cc, ll.lambda_body);

	// if no return with expression has been seen, then function
	// returns void, unless it exits with raise and cheats that
	// it returns something. So force void only if no constraints
	// present.
	if (!cc.seen_return_with_expression)
	    switch normalize_xt(ft.ret_type) {
	    case XTyVar(_):
		unify(cc, ft.ret_type, Void);
	    case _:
		skip;
	    }
	
	// now we need to insert type variables to make inferred type 
	// polimorphic
//dbgl("lambda: before " + string_of_type(ll.lambda_type));
	type_iter(subst_fv, ll.lambda_type);
	ft.tyvars = List::concat(free_vars, ft.tyvars);
//dbgl("lambda: after " + string_of_type(ll.lambda_type));
//dbgl("LAMBDA: " + string_of_type(type_map(normalize_xt, ll.lambda_type)));
    }
}

void expect(ty_ctx c, string msg, ty t, expr e)
{
    try {
	unify(c, t, type_of(e));
    } with {
    case Type_error(m):
	error("`" + msg + "' expects type " + string_of_type(t) + ", not " + 
	      string_of_type(type_of(e)));
	if (flags.verbose_errors)
	    error("(" + m + ")");
	raise Choke;
    }
}

ty tycheck_call(ty_ctx c, ty fnc, <ty>list args)
{
//dbgl("tycheck-call: " + string_of_type(fnc) + " args = (" + String::concat(", ", List::map(string_of_type, args)) + ")");
    let parms = List::map(fun *(ty, <id>option) (ty x) ((x, None)), args) in
    let ret_type = XTyVar(genxt("_ret")) in {
	try {
	    switch xt_instance(fnc) {
	    case XTyVar(_):
	    case Func(_):
		skip;
		
	    case _:
		raise Type_error("expression of non-functional type " + 
				 string_of_type(fnc) + " called");
	    }
	    
	    push_fv(c, ret_type);
	    unify(c, fnc, Func({ret_type = ret_type, 
				parms = parms, 
				tyvars = Nil}));
	} with {
	case Type_error(m):
	    error("type error in call: " + m);
	    raise Choke;
	}

//dbgl("ret: " + string_of_type(ret_type));
	return ret_type;
    }
}

ty fresh_named_type(ty_ctx c, ty_symbol ts)
{
    let tv = ts.a.tysym_tyvars in
    let t = NamedType({parms = List::map(fun (s) (TyVar(s)), tv),
		       tydef = Symbol(ts)}) in
    return fv(c, t, tv);
}

ty tycheck_union_expr(ty_ctx c, namedexpr nx)
{
    int tuple_len(ty t)
    {
	switch xt_instance(t) {
	case Tuple(lst):
	    return List::length(lst);
	case Void:
	    return 0;
	case _:
	    return 1;
	}
    }
    
    void check_with(ty t)
    {
	let ex_type = type_of(nx.nx_expr) in
	let ex_len = tuple_len(ex_type) in
	let t_len = tuple_len(t) in
	try {
	    switch (xt_instance(t), nx.nx_expr.a.raw) {
	    case (Void, VoidConst):
		skip;	// ok
	    case (Void, _):
		// we don't really want ,,void f(); Nil(f())''
		raise Type_error("void type constructor is given a value");
	    case _:
		skip;
	    }
	
	    if (ex_len != t_len) {
		let m = qualified_name(nx.nx_symbol) + " takes " + itoa(t_len) + 
		    " argument(s), not " + itoa(ex_len) in
		raise Type_error(m);
	    }
	    unify(c, ex_type, t);
	} with {
	case Type_error(m):
	    error("type error in union expression: " + m);
	    raise Choke;
	}
    }

    // first of all we typecheck the expression
    tycheck_expr(c, nx.nx_expr);
    
    // next we need to find out parent union
    var_symbol v = get_binding(nx.nx_symbol);

    if (has_property(v, IsException)) {
	check_with(v.stype);
	return Exn;
    }
    
    ty_symbol parent_union = union_of(v);

    // and finally unify type of our expression with type of
    // our union field
    let nt = fresh_named_type(c, parent_union) in
    let t = field_type_in(nt, Symbol(v)) in {
//dbgl("xtt: " +string_of_type(t));
	check_with(t);
	return nt;
    }
}

ty tycheck_struct_expr(ty_ctx c, <sini_expr>list inis)
{
    ty_symbol ts;
    
    switch inis {
    case Cons(hd, _):
	ts = (get_binding(hd.sx_symbol)).a.sf_tydef;
    case Nil:
	return ice("Type::tycheck_struct_expr(): empty sx");
    }
    
    void check_one(<sf_symbol>field fld)
    {
	id sn = simple_name(fld.fld_symbol);
	sini_expr se;

	try {
	    se = List::find(fun bool (sini_expr s) 
				(simple_name(s.sx_symbol) == sn), 
			    inis);
	} with {
	case Not_found:
	    error("struct field `" + qualified_name(fld.fld_symbol) +
		  "' not initalized");
	    return;
	}
	
	tycheck_expr(c, se.sx_expr);
	try {
//dbg("x: " + string_of_type(type_of(se.sx_expr)) + " f: " + string_of_type(fld.fld_type) + "\n");
	    unify(c, type_of(se.sx_expr), fld.fld_type);
	} with {
	case Type_error(m):
	    error("type error: " + m + " in initalization of struct field `" +
		  qualified_name(se.sx_symbol) + "'");
	}
    }
    
    let nt = fresh_named_type(c, ts) in {
	switch type_instance(nt) {
	case Struct(x):
	case OptStruct(x):
	    List::iter(check_one, x);
	    
	case _:
	    return ice("Type::tycheck_struct_expr(): non-struct");
	}

	return nt;
    }
}

ty tycheck_dot_expr(ty_ctx c, sini_expr de)
{
    tycheck_expr(c, de.sx_expr);

    let t = xt_instance(type_of(de.sx_expr)) in
    switch t {
    case Struct(_):
    case OptStruct(_):
	try {
	    _ fld = find_field_in_struct(t, de.sx_symbol);
	    
	    // Bind it. Binder couldn't do that.
	    de.sx_symbol = Symbol(get_binding(fld.fld_symbol));

	    return fld.fld_type;
	} with {
	case Not_found:
	    error("`" + simple_name(de.sx_symbol) + "' is not part of " + 
		  string_of_type(type_of(de.sx_expr)));
	    raise Choke;
	}
   
    case XTyVar(_):
	switch de.sx_symbol {
	case Symbol(sf):
	    // If binder was kind enough to set symbol, we can figure out
	    // type of this expression.
	    unify(c, fresh_named_type(c, sf.a.sf_tydef), t);
	    // And restart.
	    return tycheck_dot_expr(c, de);

	case x:
	    // otherwise we'll cry
	    error("need typing clue about `." + qualified_name(x) + "'");
	    raise Choke;
	}
	
    case _:
	error("left-hand side of `.' is of non-struct type " +
	      string_of_type(type_of(de.sx_expr)));
	raise Choke;
    }

}

// TODO: break this into smaller pieces
void tycheck_pattern(ty_ctx c, ty expr_type, pattern pat)
{
    void pat_unify(ty t, pattern p)
    {
	push_location(p.loc);
	
	try {
	    switch p.a {
	    case VoidPattern:
		skip;

	    case ExprPattern(e):
		tycheck_expr(c, e);
		try {
		    unify(c, type_of(e), t);
		} with {
		case Type_error(m):
		    error("type error with constant pattern: " + m);
		    raise Choke;
		}
		
		if (!is_constant_expr(e)) {
		    error("pattern expression is not constant");
		    raise Choke;
		}
	    
	    case VarPattern(v):
		try {
		    // it will introduce new free type variable
		    push_fv(c, v.var_type);
		    unify(c, v.var_type, t);
		} with {
		case Type_error(m):
		    error("type error with pattern variable: " + m);
		}
	    
	    case TuplePattern(lst):
		if (List::length(lst) < 2)
		    ice("Bind::tycheck_pattern(): tuple shorter then 2");

		switch xt_instance(t) {
		case Tuple(tlst):
		    try {
			List::iter2(pat_unify, tlst, lst);
		    } with {
		    case Invalid_argument("List::iter2"):
			error("number of arguments in tuple pattern and tuple " + 
			      "type doesn't match");
			raise Choke;
		    }
		    
		case XTyVar(xt):
		    let mkvar = fun (x) (XTyVar(genxt("_tuple"))) in
		    let tt = Tuple(List::map(mkvar, lst)) in
		    subst(c, xt, tt);

		    // restart
		    pat_unify(t, p);
		    
		case _:
		    error("tuple pattern used to match non-tuple type " + 
			  string_of_type(t));
		}
	    
	    case UnionPattern(n):
		let sym = get_binding(n.np_symbol) in
		if (has_property(sym, IsException)) {
		    switch xt_instance(t) {
		    case XTyVar(x):
			x.t = Some(Exn);
			pat_unify(sym.stype, n.np_pattern);
			
		    case Exn:
			pat_unify(sym.stype, n.np_pattern);
			
		    case _:
			error("this pattern matches exceptions, not " +
			      string_of_type(t));
			raise Choke;
		    }
		    
		    return;
		} 
		
		switch xt_instance(t) {
		case Union(Cons(f, _)):
		    let eu = union_of(get_binding(f.fld_symbol)) in
		    let pu = union_of(get_binding(n.np_symbol)) in
		    if (!same_symbols(eu, pu)) {
			error("pattern matching union " + symbol_name(pu) + 
			      " is used to match union " + symbol_name(eu));
			raise Choke;
		    }
		    
		    let tt = field_type_in(t, n.np_symbol) in
		    pat_unify(tt, n.np_pattern);

		case XTyVar(xt):
		    let pu = union_of(get_binding(n.np_symbol)) in
		    let nt = fresh_named_type(c, pu) in 
		    subst(c, xt, nt);
		    
		    // restart
		    pat_unify(t, p);
		    
		case _:
		    let pu = union_of(get_binding(n.np_symbol)) in {
			error("pattern matching union " + symbol_name(pu) +
			      " is used to match values of non-union type " + 
			      string_of_type(t));
			raise Choke;
		    }
		}
	    }
	} finally {
	    pop_location();
	}
    }

    pat_unify(expr_type, pat);
}

void tycheck_case_clause(ty_ctx c, ty expr_type, case_clause cc)
{
    void tycheck_one_case(*(pattern, expr) oc)
    {
	let (pat, cond) = oc in {
	    tycheck_pattern(c, expr_type, pat);
	    try {
		tycheck_expr(c, cond);
		unify(c, Bool, type_of(cond));
	    } with {
	    case Type_error(m):
		errorl(cond.loc, "case guard needs to have bool type, but " + m);
	    }
	}
    }

    List::iter(tycheck_one_case, cc.cases);
    tycheck_stmt(c, cc.body);
}

ty tycheck_binop(ty_ctx c, op2 op, expr e1, expr e2)
{
    ty t; // this is set by helper routines below

    void same_arith_types(string msg, expr ex1, expr ex2)
    {
	let t1 = xt_instance(type_of(ex1)) in
	let t2 = xt_instance(type_of(ex2)) in
	switch (t1, t2) {
	case (Int, Int):
	case (Float, Float):	
	    skip;
	    
	case (Int, Float):
	    error("`" + msg + 
		  "' won't do int -> float conversion on left argument");
	    if (flags.verbose_errors)
		error("type conversions in Gont have to be explicit");
	    t = Float;
	    
	case (Float, Int):
	    error("`" + msg + 
		  "' won't do int -> float conversion on right argument");
	    if (flags.verbose_errors)
		error("type conversions in Gont have to be explicit");
	    t = Float;
	    
	case (Int, XTyVar(_)):
	case (Float, XTyVar(_)):
	case (XTyVar(_), Float):
	case (XTyVar(_), Int):
	    unify(c, t1, t2);
	    t = type_instance(t1);
	    
	// better error message
	case (XTyVar(_), XTyVar(_)):
	    error("sorry, cannot infer types of arguments of `" + msg + "'");
	    raise Choke;
	    
	case _:
	    error("`" + msg + "' works on ints and floats, not on " +
		  string_of_type(t1) + " and " + string_of_type(t2));
	    raise Choke;
	}
    }

    void tycheck_cmp(string sop, expr ex1, expr ex2)
    {
	t = Bool;

	let t1 = xt_instance(type_of(ex1)) in
	let t2 = xt_instance(type_of(ex2)) in
	switch (t1, t2) {
	case (Int, Int):
	case (Float, Float):
	case (String, String):
	    skip;
	    
	case (Int, XTyVar(_)):
	case (Float, XTyVar(_)):
	case (String, XTyVar(_)):
	case (XTyVar(_), Int):
	case (XTyVar(_), String):
	case (XTyVar(_), Float):
	    unify(c, t1, t2);
	    
	// better error message
	case (XTyVar(_), XTyVar(_)):
	    error("sorry, cannot infer types of arguments of `" + sop + "'");
	    
	case _:
	    error("`" + sop + "' cannot work on " + string_of_type(t1) + 
		  " and " + string_of_type(t2));
	}
    }
    
    void tycheck_cmp_eq(string sop, expr ex1, expr ex2)
    {
	t = Bool;
	
	let t1 = xt_instance(type_of(ex1)) in
	let t2 = xt_instance(type_of(ex2)) in
	switch (t1, t2) {
	// null checks
	case (OptStruct(Nil), OptStruct(_)):
	case (OptStruct(_), OptStruct(Nil)):
	    skip;
	    
	// bools are ok
	case (Bool, Bool):
	case (XTyVar(_), Bool):
	case (Bool, XTyVar(_)):
	    unify(c, t1, t2);
	    
	// and other _cmp stuff:
	case _:
	    tycheck_cmp(sop, ex1, ex2);
	}
    }
    
    void tycheck_plus(expr ex1, expr ex2)
    {
	// cheat a bit, + takes the same argument types as < does
	tycheck_cmp("+", ex1, ex2);
	t = instanced_type_of(ex1);
    }
    
    tycheck_expr(c, e1);
    tycheck_expr(c, e2);

    t = type_of(e1);

    switch op {
    case Plus:	tycheck_plus(e1, e2);
    case Minus:	same_arith_types("-", e1, e2);
    case Mult:	same_arith_types("*", e1, e2);
    case Div:	same_arith_types("/", e1, e2);
    case Mod:
	expect(c, "%", Int, e1);
	expect(c, "%", Int, e2);
    case Lt:	tycheck_cmp("<", e1, e2);
    case Gt:	tycheck_cmp(">", e1, e2);
    case Ge:	tycheck_cmp(">=", e1, e2);
    case Le:	tycheck_cmp("<=", e1, e2);
    case Eq:	tycheck_cmp_eq("==", e1, e2);
    case Ne:	tycheck_cmp_eq("!=", e1, e2);
    case AndIf:
	expect(c, "&&", Bool, e1);
	expect(c, "&&", Bool, e2);
    case OrIf:
	expect(c, "||", Bool, e1);
	expect(c, "||", Bool, e2);
    case Assign:
	try {
	    unify(c, type_of(e1), type_of(e2));
	} with {
	case Type_error(m):
	    error("incompatible types in assigment: " + m);
	}

	if (!is_lvalue(e1))
	    error("assigment to non-lvalue");
    }

    return t;
}

void tycheck_raw_expr(ty_ctx c, nl_expr ex)
{
    ty self(expr e)
    {
	tycheck_expr(c, e);
	return type_of(e);
    }
    
    ty t;

    switch ex.expr_type {
    case Some(_):
	return;	    // don't bother checking it again
    case _:
	skip;
    }

    switch ex.raw {
    case SimpleBinOp(op, e1, e2):
	t = tycheck_binop(c, op, e1, e2);
	
    case SimpleUnOp(op, e):
	tycheck_expr(c, e);

	t = type_of(e);
	
	switch op {
	case Negate:
	    switch (xt_instance(type_of(e))) {
	    case Int:
		skip;
	    case Float:
		skip;
	    case _:
		error("negate operator expects integer or " +
		      "float argument, not " + string_of_type(type_of(e)));
	    }
	case PreDec:	expect(c, "pre --", Int, e);
	case PreInc:	expect(c, "pre ++", Int, e);
	case PostDec:	expect(c, "post --", Int, e);
	case PostInc:	expect(c, "post ++", Int, e);
	}
    
    case Ref(rr):
	t = (get_binding(rr.ref_symbol)).stype;

    case Call(e, el):
	t = tycheck_call(c, self(e), List::map(self, el));

    case TupleExpr(el):
	t = Tuple(List::map(self, el));

    case UnionExpr(u):
	tycheck_expr(c, u.nx_expr);
	t = tycheck_union_expr(c, u);
    
    case Lambda(l):
	tycheck_lambda(Some(c), l);
	t = l.lambda_type;

    case StructExpr(s):
	t = tycheck_struct_expr(c, s);

    case DotExpr(de):
	t = tycheck_dot_expr(c, de);
    
    case BoolConst(_):	    t = Bool;
    case IntConst(_):	    t = Int;
    case FloatConst(_):	    t = Float;
    case StringConst(_):    t = String;
    case VoidConst:	    t = Void;
    case NullConst:	    t = OptStruct(Nil);
    }

    ex.expr_type = Some(t);
}

void tycheck_expr(ty_ctx c, expr e)
{
    locate(partial(tycheck_raw_expr, c), e);
}

void warn_ignored_value(expr e)
{
    if (!flags.wignored_value)
	return;
    
    switch e.a.raw {
    case Call(_, _):
	switch xt_instance(type_of(e)) {
	case Void:
	    skip;
	case _:
	    if (flags.wignored_retval)
		warn("return value of function ignored");
	}
	
    case SimpleBinOp(Assign, _, _):
    case SimpleUnOp(PreDec, _):
    case SimpleUnOp(PostDec, _):
    case SimpleUnOp(PreInc, _):
    case SimpleUnOp(PostInc, _):
	skip;	    // ok

    case _:
	warn("value of expression ignored");
    }
}

void check_dangling_else(stmt s1, stmt s2)
{
    switch (s1.a, s2.a) {
    case (If(_, _, s3), Skip):
	switch s3.a {
	case Skip:
	    // ok, if (e1) if (e2) foo(); is ok
	    skip;   
	case _:
	    // not ok: if (e1) if (e2) foo(); else bar();
	    error("dangling else, please use `{ ... }'");
	}
	
    case _:
	// rest is ok (I hope...)
	skip;
    }
}

void non_void_type(vardef v)
{
    switch xt_instance(v.var_type) {
    case Void:
	error("variable `" + qualified_name(v.var_symbol) + 
	      "' defined with void type");
    case _:
	skip;
    }
}

void tycheck_stmt_nl(ty_ctx c, nl_stmt s)
{
    switch s {
    case Skip:
	skip;
    
    case Expr(e):
	tycheck_expr(c, e);
	warn_ignored_value(e);
    
    case Return(e):
	switch e.a.raw {
	case VoidConst:
	    skip;
	case _:
	    c.seen_return_with_expression = true;
	}
	tycheck_expr(c, e);
	unify(c, type_of(e), c.fnc_ret_type);

    case VarDef(v):
	push_fv(c, v.var_type);
	
    case Block(lst):
	List::iter(partial(tycheck_stmt, c), lst);

    case Pair(s1, s2):
	tycheck_stmt(c, s1);
	tycheck_stmt(c, s2);
    
    case Switch(e, clauses):
	tycheck_expr(c, e);
	List::iter(partial2(tycheck_case_clause, c, type_of(e)), clauses);
	
    case Let(assigns, st):
	void check_one(*(pattern, expr) a)
	{
	    let (p, e) = a in {
		tycheck_expr(c, e);
		tycheck_pattern(c, type_of(e), p);
	    }
	}

	List::iter(check_one, assigns);
	tycheck_stmt(c, st);
    
    case If(e, s1, s2):
	tycheck_expr(c, e);
	unify(c, Bool, type_of(e));
	tycheck_stmt(c, s1);
	tycheck_stmt(c, s2);
	check_dangling_else(s1, s2);
    
    case Loop(l):
	tycheck_stmt(c, l.ll_body);
	tycheck_stmt(c, l.ll_inc);
    
    case Break(_):
	skip;
	
    case BreakIfNot(e):
	tycheck_expr(c, e);
	unify(c, Bool, type_of(e));
    
    case Open(_):
	skip;
    
    case Try(tr):
	tycheck_stmt(c, tr.try_body);
	List::iter(partial2(tycheck_case_clause, c, Exn), tr.try_with);
	tycheck_stmt(c, tr.try_finally);
    
    case Raise(e):
	tycheck_expr(c, e);
	unify(c, Exn, type_of(e));
    }
}

void tycheck_stmt(ty_ctx ctx, stmt s)
{
    void check(nl_stmt ss)
    {
	try {
	    tycheck_stmt_nl(ctx, ss);
	} with {
	case Type_error(m):
	    error(m);
	case Choke:
	    skip;   // just jump to next stmt
	}
    }
    
    locate(check, s);
}

void check_var_with_interface(var_symbol impl, string extern_name)
{
    var_symbol iface;

    if (extern_name != "")
        set_property(impl, IsExported);
    
    try {
	iface = Module::lookup_var(impl.module, impl.name);
    } with {
    case Not_found:
	return;
    }

    if (extern_name != "") {
	error("function defined with `extern_as' and declared in interface");
	return;
    }
    
    try {
	unify(dummy_ty_ctx(), iface.stype, impl.stype);
    } with {
    case Type_error(s):
	error("implementation of " + symbol_name(impl) + 
	      " is incompatible with type declared in interface: " + s);
	errorl(iface.sloc, "type in interface is: " + 
	       string_of_type(iface.stype));
    }

    // set this for Codegen, so it will know to put (public) in front of it
    set_property(impl, IsExported);
    // and this for ourselves so we can check whatever all variables
    // declared were defined
    set_property(iface, IsExported);
}

void check_module_completness()
{
    if (Bind::doing_iface)
	return;
    
    Module::module m = Module::open_module(Bind::current_module);

    <var_symbol>list missing = Nil;

    void checker(string dummy, var_symbol s)
    {
	if (!has_property(s, IsExported) &&
	    !has_property(s, IsUnionField) &&
	    !has_property(s, IsExternC))
	    missing = Cons(s, missing);
    }

    Hashtbl::iter(m.mod_vars, checker);

    if (List::length(missing) != 0) {
	error("symbols declared in interface but not implemented:");
	void print(var_symbol s)
	{
	    errorl(s.sloc, "- `" + s.name + "' of type " +
		   string_of_type(s.stype));
	}
	List::iter(print, missing);
    }
}

void tycheck_nl_decl(nl_decl nld)
{
    switch nld {
    case FuncDecl(fn):
	let s = get_binding(fn.fn_symbol) in {
	    check_var_with_interface(s, fn.fn_export_as);
	    // Use bogus context, typcheck_lambda() is expected
	    // to make its own.
	    tycheck_lambda(None, fn.fn_lambda);
	}
    
    case SectionDecl(_, l):
	// likewise
	tycheck_lambda(None, l);
    
    case TopLevelOpen(_):
	skip;
	
    case GlobalDecl(v):
	if (!Bind::doing_iface)
	    check_var_with_interface(get_binding(v.var_symbol), "");
	non_void_type(v);
	
    case ExternC(v):
	non_void_type(v);
	
    case TyDef(_):
	// redefinition and compatibility with interface was already
	// checked by binder
	skip;
    
    case FuncProto(_):
    case XCFuncProto(_, _):
    case ExceptionDecl(_):
    case DummyDecl:
	skip;
    }
}

void tycheck_decl(decl d)
{
    locate(tycheck_nl_decl, d);
}

// vim:sts=4:sw=4
