/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 702 $ $LastChangedDate: 2002-08-30 01:19:26 +0200 (Fri, 30 Aug 2002) $
 *
 * Util - Various utility functions: diagnostics, compilation flags, etc.
 */
 
open Ast;

/* Prints message and sets bailout flag.  The flag is tested by maybe_choke()
   routine.  */
void errorl(location loc, string msg);

/* Prints message prepended with 'warning: ', sets bailout flag is -Werror
   was specified in command line.  */
void warnl(location loc, string msg);

/* If bailout flag is set raise Exit.  */
void maybe_choke();

/** Location stack. 

    Whenever we start processing some language construct with location
    included in Ast, we push its location, on stack.  After we are finish
    with it, we pop it.  This allows various routines just to call
    error()/warning() functions without any location, and have the location
    automagically prepended.  */

/* Push location on the stack.  */
void push_location(location loc);

/* Pop it.  */
void pop_location();

/* Return topmost location of the stack. If stack is empty we return last
   location pushed, no matter if it was popped.  This is used in constructions
   of artificial Ast's.  */
location top_location();

/* Print warning message in topmost location context.  */
void warn(string msg);

/* Print error message in topmost location context.  */
void error(string msg);

/* Run [f(x.a)] with [x.loc] pushed before and popped after.  */
'a locate(*('a ('b)) f, <'b>located x);

/* Prints error message and does abort().  ICE == Internal Compiler Error.  
   It uses top_location() so user can figure out how to code it around.
   If bailout flag is set, then we print "confused by earlier errors,
   bailing out", otherwise "internal compiler error".  */
'a ice(string msg);

/* Returning <'a>symbol for given binding.  If not bound -- ICE.  */
'a get_binding(<'a>bindable sym);

/* Compiler flags.  */
struct compiler_flags {
	bool werror;
	bool wshadow;
	bool wdead;
	bool wlabeled;
	bool wpattern;
	bool wcapitalization;
	bool wignored_value;
	bool wignored_retval;
	bool verbose_errors;
	bool time_report;
}

/* Current compiler settings.  */
compiler_flags flags;

/** Debugging utilities.  */

/* Same as [print_string(m)]. */
void dbg(string m);

/* Like [dbg()] but also prepend line number and append newline.  */
void dbgl(string s);

/* Partially evaluate [f] on [c].  Well... this is poor, it should be some
   language construct for it probably... anyway now when we have type 
   inference, we [partial(c, f)] can be written as [fun (x) (f(c, x))] which
   is about the same length, but I guess less readable.  */
*('a ('b)) partial(*('a ('c, 'b)) f, 'c c);

/* Same as partial(), but partially evaluate two arguments.  */
*('a ('b)) partial2(*('a ('c, 'd, 'b)) f, 'c c, 'd d);

/* Return [true] iff two [s1] and [s2] are the same symbol.  */
bool same_symbols(<'a>symbol s1, <'a>symbol s2);

/* Return unique name starting with [prefix].  */
string genname(string prefix);

/* Apply [f] to [t], and the recursively to each component type of
   what it returns.  For example [f] return Struct(...), then it's
   later applied to each [fld_type] and such mapped structure is
   returned.  It does not walk down XTyVar(...) chains.  */
ty type_map(*(ty (ty)) f, ty t);

/* Similar to [type_map()], but [f] can only modify [t] in-place, not return
   new copy of it.  It also differs from [type_map()] in that it walks down
   XTyVar(...) chains.  */
void type_iter(*(void (ty)) f, ty t);

/* Return fresh copy of [t].  Used in making Ast of [_ x, y, z;] where
   each of [x, y, z] gets it's own copy of weak type.  */
ty copy_type(ty t);

/* Add [p] to [s] properties list.  */
void set_property(var_symbol s, var_symbol_property p);

/* Check whether [s] has property [p].  Only tag of union is checked, so
   [has_property(s, ParentUnion(x))] will be [true] is [s] has any
   [ParentUnion(_)] property, not specifically [ParentUnion(x)].  */
bool has_property(var_symbol s, var_symbol_property p);

/* Return just name component of symbol.  */
string simple_name(<<'a>symbol>bindable s);

/* Return [module::name] form name. If symbol name includes no module -- return
   just [name].  */
string qualified_name(<<'a>symbol>bindable s);

/* Decomposite [module::name] form name into QualifiedName(...) or 
   SimpleName(...).  Used by module serializer.  */
<<'a>symbol>bindable rev_qualified_name(string name);

/* Same as [qualified_name(Symbol(x)].  */
string symbol_name(<'a>symbol x);

/* Raised by routines that cannot return anything because of user error.
   It might be cought if value returned by function in question is not needed
   at some higher level.  If it gets to main, compilation terminates.  */
exception void Choke;

/* Raised when we really want to get out the compiler.  */
exception void Bail_out;
