/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 686 $ $LastChangedDate: 2002-08-29 21:14:56 +0200 (Thu, 29 Aug 2002) $
 *
 * Xtcheck - check whether there are none weak type variables.
 */

open Ast;
open Util;
open Tyutil;

void subst_dummies(ty t)
{
    void f(ty tt)
    {
	switch tt {
	case XTyVar(x):
	    switch x.t {
	    case None:
		x.t = Some(TyVar("$dummy"));
	    case Some(_):
		skip;
	    }
	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

bool xtcheck_type(ty t)
{
    bool has_weak = false;
    
    void f(ty tt)
    {
	switch tt {
	case XTyVar(x):
	    switch x.t {
	    case None:
		has_weak = true;
	    case Some(_):
		skip;
	    }
	case TyVar("$dummy"):
	    has_weak = true;
	case _:
	    skip;
	}
    }

    type_iter(f, t);
    
    if (has_weak)
	error("weak type variables not allowed here (type is " + 
		  string_of_type(t) + ")");

    return has_weak;
}

void xtcheck_raw_expr(nl_expr ex)
{
    switch ex.raw {
    case SimpleBinOp(_, e1, e2):
	xtcheck_expr(e1);
	xtcheck_expr(e2);
	
    case SimpleUnOp(_, e):
	xtcheck_expr(e);
    
    case Ref(_):
	skip;

    case Call(e, el):
	xtcheck_expr(e);
	List::iter(xtcheck_expr, el);

    case TupleExpr(el):
	List::iter(xtcheck_expr, el);

    case UnionExpr(u):
	xtcheck_expr(u.nx_expr);
    
    case Lambda(l):
	xtcheck_lambda(l);

    case StructExpr(ss):
	List::iter(fun void (sini_expr s) (xtcheck_expr(s.sx_expr)), ss);

    case DotExpr(d):
	xtcheck_expr(d.sx_expr);
    
    case BoolConst(_):
    case IntConst(_):
    case FloatConst(_):
    case StringConst(_):
    case VoidConst:
    case NullConst:
	skip;
    }

    switch ex.expr_type {
    case Some(t):
	subst_dummies(t);
	
    case None:
	ice("Xtcheck::xtcheck_raw_expr(): no type?");
    }
}

void xtcheck_expr(expr e)
{
    locate(xtcheck_raw_expr, e);
}

void xtcheck_nl_pattern(nl_pattern p)
{
    switch p {
    case VoidPattern:
	skip;
	
    case VarPattern(v):
	if (xtcheck_type(v.var_type))
	    error("in pattern variable " + simple_name(v.var_symbol));
    
    case ExprPattern(e):
	xtcheck_expr(e);

    case TuplePattern(lst):
	List::iter(xtcheck_pattern, lst);
    
    case UnionPattern(np):
	xtcheck_pattern(np.np_pattern);
    }
}

void xtcheck_pattern(pattern p)
{
    locate(xtcheck_nl_pattern, p);
}

void xtcheck_nl_stmt(nl_stmt s)
{
    void xtcheck_case_clause(case_clause cc)
    {
	List::iter(xtcheck_pattern, List::map(fst, cc.cases));
	List::iter(xtcheck_expr, List::map(snd, cc.cases));
	xtcheck_stmt(cc.body);
    }
    
    switch s {
    case Skip:
	skip;
    
    case Expr(e):
	xtcheck_expr(e);
    
    case Return(e):
	xtcheck_expr(e);

    case VarDef(v):
	if (xtcheck_type(v.var_type))
	    error("in variable definition of " + simple_name(v.var_symbol));
	else
	    switch type_instance(v.var_type) {
	    case Void:
		// it is only allowed for parameters
		error("variable `" + simple_name(v.var_symbol) +
		      "' is bound to have `void' type");
	    case _:
		skip;
	    }
	
	
    case Block(lst):
	List::iter(xtcheck_stmt, lst);

    case Pair(s1, s2):
	xtcheck_stmt(s1);
	xtcheck_stmt(s2);
    
    case Switch(e, clauses):
	xtcheck_expr(e);
	List::iter(xtcheck_case_clause, clauses);
	
    case Let(assigns, st):
	void check_one(*(pattern, expr) a)
	{
	    let (p, e) = a in {
		xtcheck_expr(e);
		xtcheck_pattern(p);
	    }
	}

	List::iter(check_one, assigns);
	xtcheck_stmt(st);
    
    case If(e, s1, s2):
	xtcheck_expr(e);
	xtcheck_stmt(s1);
	xtcheck_stmt(s2);
    
    case Loop(l):
	xtcheck_stmt(l.ll_body);
	xtcheck_stmt(l.ll_inc);
    
    case Break(_):
	skip;
	
    case BreakIfNot(e):
	xtcheck_expr(e);
    
    case Open(_):
	skip;
    
    case Try(tr):
	xtcheck_stmt(tr.try_body);
	List::iter(xtcheck_case_clause, tr.try_with);
	xtcheck_stmt(tr.try_finally);
    
    case Raise(e):
	xtcheck_expr(e);
    }
}

void xtcheck_stmt(stmt s)
{
    locate(xtcheck_nl_stmt, s);
}

void xtcheck_lambda(lambda_record l)
{
    List::iter(fun void (var_symbol s) {
		    if (xtcheck_type(s.stype))
			error("in lambda parameter " + s.name);
		}, 
	       l.lambda_parms);
    if (xtcheck_type(l.lambda_type))
	error("in lambda expression type");
    xtcheck_stmt(l.lambda_body);
}

void xtcheck_nl_decl(nl_decl nld)
{
    switch nld {
    case FuncDecl(fn):
	xtcheck_lambda(fn.fn_lambda);
    
    case SectionDecl(_, l):
	xtcheck_lambda(l);
    
    case GlobalDecl(v):
    case ExternC(v):
    case FuncProto(v):
    case XCFuncProto(_, v):
    case ExceptionDecl(v):
	if (xtcheck_type(v.var_type))
	    error("in global symbol " + qualified_name(v.var_symbol));
	
    case TyDef(ts):
	if (xtcheck_type(ts.stype))
	    error("in type definition");
	
    case TopLevelOpen(_):
    case DummyDecl:
	skip;
    }
}

void xtcheck_decl(decl d)
{
    locate(xtcheck_nl_decl, d);
}

// vim:sts=4:sw=4
